"""cssutils lexer

TODO
    quotes: included \"?
    replace blocks!
    simplify

"""
__docformat__ = 'restructuredtext'
__version__ = '0.8a5'

import re
import types
import xml.dom

import cssutils


class CSSStyleSheetLexer(object):


    def __init__(self):
        self.log = cssutils._log
        

    # quotetext " " or ' '
    _requotes = re.compile(ur''' " .*? " | ' .*? ' ''',
                           re.DOTALL|re.UNICODE|re.VERBOSE)

    def _replaceQuotes(self, cssText):
        """
        replaces all Quotes in cssText with simple quotes in
        format "idofquote" or 'idofquote'

        returns replacedText        
        """
        self._doublequotes = []
        self._singlequotes = []
        
        d = s = 0
        for t in self._requotes.findall(cssText):
            if t.startswith(u'"'):
                self._doublequotes.append(t)
                cssText = cssText.replace(t, u'"%s"' % d)
                d += 1
            elif t.startswith(u"'"):
                self._singlequotes.append(t)
                cssText = cssText.replace(t, u"'%s'" % s)
                s += 1

        return cssText


    def _restoreQuotes(self, cssText):
        """
        replaces all Quotes in cssText with original quote content

        returns replacedText
        """       
        d = s = 0
        for t in self._requotes.findall(cssText):
            index = int(t[1:-1])
            if t.startswith(u'"'):
                cssText = cssText.replace(t, self._doublequotes[index])
                d += 1
            elif t.startswith(u"'"):
                cssText = cssText.replace(t, self._singlequotes[index])
                s += 1

        return cssText


    # comment /* ... */
    _recomments = re.compile(ur'(\/\*.*?\*\/)', re.DOTALL|re.UNICODE)

    def _replaceComments(self, cssText):
        """
        replaces all comments in cssText with simple comments in
        format /*idofcomment*/

        returns replacedText        
        """
        newcss = []
        self._comments = []
        
        i = 0
        for t in self._recomments.split(cssText):
            if t.startswith(u'/*'):
                self._comments.append(t)
                newcss.append(u'/*%s*/' % i)
                i += 1
            else:
                newcss.append(t)

        return u''.join(newcss)   


    def _restoreComments(self, cssText):
        newcss = []
        
        for t in self._recommentsonly.split(cssText):
            if t.startswith(u'/*'):
                i = int(t[2:-2])
                newcss.append(self._comments[i])
            else:
                newcss.append(t)

        return u''.join(newcss)


    def _preprocess(self, cssText):
        """
        - removes "\\t" and "\\f" whitespace
        - replaces all kinds of newlines ("\\n", "\\r", "\\r\\n") to "\\n"
        """
        # comments
        cssText = self._replaceComments(cssText)
        # quotes
        cssText = self._replaceQuotes(cssText)
        # remove HTML comment delimiters
        cssText = cssText.replace(u'<!--', u'').replace(u'-->', u'')
        # WS normalize
        cssText = cssText.replace(u'\r\n', u'\n').replace(u'\r', u'\n').replace(u'\f', u'\n')
        cssText = cssText.replace(u'\t', u' ')

        return cssText


    # comment /* ... */ only 
    _recommentsonly = re.compile(ur'(\/\*.*?\*\/)', re.DOTALL|re.UNICODE)

    def _postprocess(self, *cssTexts):
        """
        restores comments and quotes
        
        returns tuple of mergedcssTexts or
        if input was single cssText returns this only
        """
        restored = []

        for cssText in cssTexts:        
            # quotes
            cssText = self._restoreQuotes(cssText)
            # comments
            cssText = self._restoreComments(cssText)

            restored.append(cssText)
            
        if len(restored) == 1:
            return restored[0] # 1 unicodestring
        else:
            return tuple(restored) # tuple of unicodestrings


    # --- CHARSET ---
    _reIsCSSCharSetRule = re.compile(
        ur'''@charset (.*?)''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                
    
    _reCSSCharsetRule = re.compile(
        ur'''@charset (.*?) ("[^"]*?") (.*?) ;''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    def lexCSSCharsetRule(self, cssText, line):
        """
        returns (postC, encoding, postC)
        """
        cssText = self._preprocess(cssText)

        if not self._reIsCSSCharSetRule.match(cssText):
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Expected CSSCharsetRule but found: %s' %
                           cssText, line, xml.dom.InvalidModificationErr)
            return 3 * (None,)

        try:
            preC, encoding, postC =\
                self._reCSSCharsetRule.findall(cssText)[0]
        except IndexError:
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Invalid CSSCharsetRule: %s' %
                           cssText, line, xml.dom.SyntaxErr)
            return 3 * (None,)

        preC, encoding, postC = self._postprocess(
            preC, encoding, postC)

        encoding = encoding[1:-1]

        return (preC, encoding, postC)

    
    # --- IMPORT ---

    _reIsCSSImportRule = re.compile(
        ur'''
        @import (.*?)''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    _reCSSImportRule = re.compile(
        ur'''@import (.*?) url\(([^"']*?)\) (.*?) ;''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    def lexCSSImportRule(self, cssText, line):
        """
        returns (hrefC, href, media)
            @import [x1comment] url(...) mediainclcomments
        """
        # TODO: ImportRule comments

        cssText = self._preprocess(cssText)

        if not self._reIsCSSImportRule.match(cssText):
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Expected CSSImportRule but found: %s' %
                           cssText, line, xml.dom.InvalidModificationErr)
            return 3 * (None,)

        try:
            hrefC, href, media = self._reCSSImportRule.findall(
                cssText)[0]
        except IndexError:
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Invalid CSSImportRule: %s'
                           % cssText, line, xml.dom.SyntaxErr)
            return 3 * (None,)

        return self._postprocess(hrefC, href, media)


##    # --- MEDIARULE ---
##
##    _reIsCSSMediaRule = re.compile(
##        ur'''@media (.*?)''',
##        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                
##
##    _reCSSMediaRule = re.compile(
##        ur'''@media (.*?) \{ (.*?) \} \s*''',
##        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)            
##
##    def lexCSSMediaRule(self, cssText, line):
##        """
##        returns mediaText, rulesText
##        both inclusive comments
##        """
##        cssText = self._preprocess(cssText)
##
##        if not self._reIsCSSMediaRule.match(cssText):
##            cssText = self._postprocess(cssText)
##            self.log.error(u'(L) Expected CSSMediaRule but found: %s' %
##                           cssText, line, xml.dom.InvalidModificationErr)
##              return ...       
##        try:
##            mediaText, rulesText =\
##                 self._reCSSMediaRule.findall(cssText)[0]          
##        except (IndexError):
##            cssText = self._postprocess(cssText)
##            self.log.error(u'(L) Invalid CSSMediaRule: %s' %
##                           cssText, line, xml.dom.SyntaxErr)
##            return...
##
##        return self._postprocess(mediaText, rulesText)

        
    # --- STYLERULE ---

    _reIsCSSStyleRule = re.compile(
        ur'''^[^\@].*? { .*''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    _reCSSStyleRule = re.compile(
        ur'''(.+?) ( \{ .*? \} ) \s*''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)            

    def lexCSSStyleRule(self, cssText, line):
        """
        returns selectorText, styledeclarationText
        both inclusive comments
        """
        cssText = self._preprocess(cssText)

        if not self._reIsCSSStyleRule.match(cssText):
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Expected CSSStyleRule but found: "%s"' %
                           cssText, line, xml.dom.InvalidModificationErr)
            return 2 * (None,)
        
        try:
            selectorText, styledeclarationText =\
                 self._reCSSStyleRule.findall(cssText)[0]          
        except (IndexError):
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Invalid CSSStyleRule: %s' %
                           cssText, line, xml.dom.SyntaxErr)
            return 2 * (None,)

        selectorText, styledeclarationText = self._postprocess(
            selectorText, styledeclarationText)

        return selectorText, styledeclarationText
        

    # --- STYLEDECLARATION ---

    _reCSSStyleDeclaration = re.compile(ur'''
        {?                # optional start {
        ( .*? )
        (?:
            ;             # end of Property ;
            |             # OR
            ;? \s* }? $   # end of StyleDeclaration ;? }? $
        )
        \s*
        ''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    def lexCSSStyleDeclaration(self, cssText, line):
        """
        returns list of Properties including
        comments            
        """
        cssText = self._preprocess(cssText)

        if not self._reCSSStyleDeclaration.match(cssText):
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Invalid CSSStyleDeclaration: %s' %
                           cssText, line, xml.dom.SyntaxErr)
            return []
        
        props = self._reCSSStyleDeclaration.findall(cssText)
        
        properties = []
        for p in props:
            if p: # should not happen but re seems not completely right
                properties.append(self._postprocess(p))

        return properties


    # --- PROPERTY---
    _reCSSProperty = re.compile(ur'''
        ([^;]*?)     # name
        :            # :
        (.*?)        # value
        (\!.*?)?     # priority
        ;?           # ;
        $''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    def lexCSSProperty(self, cssText, line):
        """
        returns (name, value, priority) including
        comments
        """        
        if not cssText:
            self.log.debug(
                u'(L) Empty Property on line %s' % line)
            return 3 * (None,)
        
        if not self._reCSSProperty.match(cssText):
            self.log.error(
                u'(L) Invalid Property: "%s" on line %s'
                % (cssText, line))
            return 3 * (None,)

        cssText = self._preprocess(cssText)

        try:
            name, value, priority =\
                self._reCSSProperty.findall(cssText)[0]
        except IndexError:
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Invalid Property: %s' %
                           cssText, line, xml.dom.SyntaxErr)
            return 3 * (None,)

        priority = priority.lower()
        
        return self._postprocess(name, value, priority)


    _reCSSPropertyName = re.compile(ur'''
        ( [^/\s]+ )|      # name
        (\/\* .*? \*\/)|  # comment
        \s*
        ''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    def lexCSSPropertyName(self, cssText, line):
        """
        returns list of comments and name
        """               
        cssText = self._preprocess(cssText)
        order = [ self._postprocess(x)
                  for x in self._reCSSPropertyName.split(cssText)
                  if x]
        return order


    # --- STYLESHEET ---

    _reCSSStyleSheet = re.compile(ur'''
        (   
            @.+?                  # @ charset|import|page|UNKNOWN
            (?:;)                 # ; or { TODO: block }
        )|
        (   
            [\w\*#.:]+? .*?            # ruleset selector,* 
            { .*? }               # { declaration,* }
        )|
        (   # Comment
            # \/\* [^*]* \*+ ([^/*][^*]*\*+)* \/ # CSS syntax
            /\* \d+ \*/          # Comment (only digits after replacement)
        )|
        (   
            \s*                   # WS
        )

##        (   ### MEDIARULE
##            (?: \/\* .*? \*\/ | \s)*    # WS or comment
##            @media .+? \{               # @media medium* { 
##            (?: .*? \{ .*? \} )*         # ruleset
##            \}                           # }
##            \s*                         # WS
##        )|  ### OR
##        (   ### MEDIARULE-END?
##            [}]
##        )|  ### OR
        ''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)            
    """
    splits rules of a stylesheet
    """

    _reWS = re.compile(ur'^\s*$', re.IGNORECASE|re.UNICODE)

    _reFindnextAtrule = re.compile(ur'@ (charset|import|page)', #media!
                                   re.IGNORECASE|re.UNICODE|re.VERBOSE)
    """matches @... (e.g. @import) to be used to _atrulenames as key"""

    _atrulenames = { u'charset': u'CSSCharsetRule',
                    u'import': u'CSSImportRule',
##                    u'media': u'CSSMediaRule',
                    u'page': u'CSSPageRule'}
    """mapping of @... to @Rulename"""

    def lexCSSStyleSheet(self, cssText, line=1):
        """
        Returns tuples (Rulename, cssText, line)
        """
        cssText = self._preprocess(cssText)

        rules = []
        self.log.debug(u'DQUOTES : %s' %
                       self._doublequotes, neverraise=True)
        self.log.debug(u'SQUOTES : %s' %
                       self._singlequotes, neverraise=True)
        self.log.debug(u'COMMENTS: %s' % self._comments, neverraise=True)
        self.log.debug(u'CSSTEXT :\n%s\n' % cssText, neverraise=True)
        self.log.debug(u'TOKENS  : %s\n' % self._reCSSStyleSheet.split(
            cssText), neverraise=True)

        for cssText in [r for r in self._reCSSStyleSheet.split(cssText)
                         if r]:
            # which rule?
            try:
                atrule = self._reFindnextAtrule.findall(cssText)[0]
                rulename = self._atrulenames.get(atrule, u'CSSUnknownRule')
            except IndexError:
##                # @mediaend a single }
##                if cssText == u'}':
##                    rulename = u'BLOCK-END'
                if cssText.startswith(u'/*'):
                    rulename = u'Comment'
                elif self._reWS.match(cssText):
                    rulename = u'WS'
                else:
                    rulename = u'CSSStyleRule'

            cssText = self._postprocess(cssText)
            rules.append( (rulename, cssText, line) )
            line += cssText.count(u'\n')

        self.log.debug(u'RULES   : %s\n' % rules,
                       neverraise=True)
        return rules
       
    
    # --- UNKNOWNRULE ---
    _knownAtRules = [u'charset', u'import'] #u'media', u'page']

    _reIsCSSUnknownRule = re.compile(
        ur'''^@\w+ (.*?) ({ .*? }$) | (;?$)''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    _reCSSUnknownRule = re.compile(
        ur'''
        ^@([_a-zA-Z]+[a-zA-Z0-9-]*)
        \s*?
        ([^;{}]*?)
        \s*?
        (
            \;$
            |
            (?:{ . *? })$
        )
        ''', 
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    def lexCSSUnknownRule(self, cssText, line):
        """
        returns (name, content, block)
        TODO: comments
        """
        if not cssText.startswith('@'):
            self.log.error(u'(L) Expected CSSUnknownRule but found: "%s"' %
                           cssText, line, xml.dom.InvalidModificationErr)
            return 3 * (None,)

        cssText = self._preprocess(cssText)
      
        try:
            name, content, semiOrBlock =\
                self._reCSSUnknownRule.findall(cssText)[0]
        except IndexError:
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Invalid CSSUnknownRule: "%s"' %
                           cssText, line, xml.dom.SyntaxErr)        
            return 3 * (None,)

        if semiOrBlock == u';':
            block = None
        else:
            block = semiOrBlock

        name, content, semiOrBlock = self._postprocess(
            name, content.strip(), semiOrBlock)

        if name in self._knownAtRules:
            cssText = self._postprocess(cssText)
            self.log.error(u'(L) Expected CSSUnknownRule but found: "%s"' %
                           cssText, line, xml.dom.InvalidModificationErr)
            return 3 * (None,)
        
        return (name, content, block)


    # --- MEDIALIST ---

    _reMediaListItem = re.compile(ur'''
        \s*
        (\/\* .*? \*\/)* # comment
        \s*
        ([a-z]+)?        # 1 medium
        \s*
        (\/\* .*? \*\/)* # comment
        \s*''',
        re.DOTALL|re.IGNORECASE|re.UNICODE|re.VERBOSE)                

    _reMedium = re.compile(ur'^[a-z]+$', re.IGNORECASE|re.UNICODE)

    _reComment = re.compile(ur'^\/\*.*?\*\/$',
                            re.DOTALL|re.IGNORECASE|re.UNICODE)

    def lexMediaList(self, mediaText, line):
        """
        returns list of media and x
        """
        # TODO: lexMediaList comments
        
        mediaText = self._preprocess(mediaText)
        
        media = []
        for medium in mediaText.split(u','):

            # check if valid
            if not self._reMediaListItem.match(medium):
                medium = self._postprocess(medium)
                self.log.error(u'(L) Invalid Medium: %s' %
                               medium, line, xml.dom.SyntaxErr)
                continue
            
            nodes = [n for n in self._reMediaListItem.split(medium)
                     if n]

            mediumdone = False
            for n in nodes:
                if not mediumdone and self._reMedium.match(n):
                    med = self._postprocess(n)
                    media.append(med)
                    mediumdone = True
                elif mediumdone and self._reMedium.match(n):
                    medium = self._postprocess(medium)
                    self.log.error(u'(L) Invalid Medium: %s' %
                               medium, line, xml.dom.SyntaxErr)
                    continue
                elif self._reComment.match(n):
                    n = self._postprocess(n)
                    self.log.debug(
                        u'TODO: lexMediaList comment %s' % n)
        return media


if __name__ == '__main__':
    import logging
    import pprint
    import sys

    cssutils._log.setloglevel(logging.DEBUG)

    L = CSSStyleSheetLexer()
    tokens = L.lexCSSStyleSheet(open(sys.argv[1]).read())
    pprint.pprint(tokens)
    