"""
Parser implementation used by CSSParser, css and stylesheets classes

"""
__docformat__ = 'restructuredtext'
__all__ = ['_getParser']
__version__ = "0.8a5"


import codecs
import xml.dom

import _lexer as lexer
import _valuevalidator as valuevalidator

from css import comment
from css import cssstylesheet
from css import cssrulelist
from css import csscharsetrule
from css import cssimportrule
## from css import cssmediarule
from css import cssstylerule
from css import cssunknownrule

import cssutils


class Parser(object):
    """
    methods to parse various parts of a CSS style sheet
    """

    _VALID_PROPERTIES = valuevalidator.cssvalues
    """list of valid properties and its values"""

    _rules = {
        u'Comment': comment.Comment,
        u'CSSCharsetRule': csscharsetrule.CSSCharsetRule,
        u'CSSImportRule': cssimportrule.CSSImportRule,
        ## u'CSSMediaRule': cssmediarule.CSSMediaRule,
        u'CSSUnknownRule': cssunknownrule.CSSUnknownRule,
        u'CSSStyleRule': cssstylerule.CSSStyleRule
        }


    def __init__(self):
        self.log = cssutils._log
        self._L = lexer.CSSStyleSheetLexer() 


    def parseCommentRule(self, cssText, line=None):
        """
        checks
        - if its a Comment at all
        - if Comment contains "/*" or "*/" which are not allowed

        returns cssText or None        
        """
        # TODO: should not happen?
        if not cssText:
            self.log.debug(u'(P) Expected Comment but nothing found',
                           line, xml.dom.SyntaxErr)
            return None
        if not (cssText.startswith(u'/*') and cssText.endswith(u'*/')):
            self.log.error(u'(P) Expected Comment but found: "%s"' %
                           cssText, line, xml.dom.InvalidModificationErr)
            return None
        
        if cssText[2:-2].find(u'/*') != -1 or\
           cssText[2:-2].find(u'*/') != -1:
            self.log.error(
                u'(P) Comment content MUST NOT contain "/*" or "*/": %s' %
                cssText)
            return None
        
        return cssText


    def parseCSSCharsetRule(self, cssText, line=None):
        return self._L.lexCSSCharsetRule(cssText, line)

    
    def parseCSSImportRule(self, cssText, line=None):
        return self._L.lexCSSImportRule(cssText, line)


    def parseCSSMediaRule(self, cssText, line=None):
        return self._L.lexCSSMediaRule(cssText, line)
    

    def parseCSSUnknownRule(self, cssText, line=None):
        return self._L.lexCSSUnknownRule(cssText, line)        


    def parseMediaList(self, mediaText, line=None):
        return self._L.lexMediaList(mediaText, line)


    def parseCSSStyleRule(self, cssText, line=None):
        return self._L.lexCSSStyleRule(cssText, line)
        

    def parseCSSStyleDeclaration(self, cssText, line=None):
        return self._L.lexCSSStyleDeclaration(cssText, line)


    def parseCSSProperty(self, cssText, line=None):
        return self._L.lexCSSProperty(cssText, line)


    def parsePropertyName(self, cssText, line=None):
        """
        fills list with (type, value) pairs
        should be 1 type "name" with optional "C" (comments)
        returns (name, listofnamecomments)"""
        name = None
        namedorder = []
        found = 0

        for x in self._L.lexCSSPropertyName(cssText, line):
            if x.startswith(u'/*'):
                namedorder.append((u'C', x))
            else:
                namedorder.append((u'name', x))
                name = x
                found += 1

        if found <> 1:
            self.log.error(
                u'(P) Invalid Propertyname: "%s"' % cssText, line)
            return None, []

        return name, namedorder


    def parsePropertyValue(self, name, cssText, line=None):
        """returns value"""
        # TODO: lex?
        # TODO: should raise only if syntax err else log!

        value = cssText.strip()
        
        if self._VALID_PROPERTIES.has_key(name):
            if not self._VALID_PROPERTIES[name](value):
                self.log.warn(
                    u'Invalid property value "%s" in CSSStyleDeclaration'
                      % cssText, line)
        else:
            self.log.warn(
                u'Unrecognized property "%s" with value "%s" in CSSStyleDeclaration'
                  % (name, value), line, neverraise=True)

        return value


    def parsePropertyPriority(self, cssText, line=None):
        """returns priority"""
        # TODO: lex?
        if not cssText == u'!important' and not cssText == u'':
            self.log.warn(
                u'Invalid property priority "%s" in CSSStyleDeclaration'
                  % cssText, line, xml.dom.SyntaxErr)
        return cssText


    
    def parseCSSStyleSheet(self, cssText, line=1):
        """returns CSSRuleList"""
        # TODO!!!
        rules = cssrulelist.CSSRuleList()

        nomorecharset = False
        nomoreimport = False

        tokens, i = self._L.lexCSSStyleSheet(cssText, line), 0
        while i < len(tokens):
            name, cssText, line  = tokens[i]
            i += 1

            if u'WS' == name:
                line += cssText.count(u'\n')
                continue

            # do complete @media including all rules inside!            
            elif u'CSSMediaRule' == name:
                nomorecharset = True
                nomoreimport = True

                rule = self._rules[name](startline=line)
                # add missing end of media rule
                cssText += u'}'
                try:
                    rule._setCssText(cssText, self)
                except xml.dom.SyntaxErr, e:
                    self.log.error(u'(P) %s' % e,
                        line, xml.dom.SyntaxErr)
                    continue                

                mediarules = cssrulelist.CSSRuleList()
                while i < len(tokens):
                    name, cssText, line = tokens[i]
                    i += 1
                    
                    if u'BLOCK-END' == name:
                        break
                    elif u'CSSStyleRule' == name: # maybe other?
                        r = self._rules[name](startline=line)
                        try:
                            r._setCssText(cssText, self)
                        except xml.dom.SyntaxErr, e:
                            self.log.error(u'(P) %s' % e,
                                line, xml.dom.SyntaxErr)
                            continue                
                
                    if r:
                        mediarules.append(r)

                rule.cssRules = mediarules

            elif u'BLOCK-END' == name:
                self.log.error(
                    u'(P) Found end of Rule Block "%s" but no open block'
                    % cssText, line, xml.dom.HierarchyRequestErr)
                continue

            # check hierarchy
            elif u'CSSCharsetRule' == name:
                if nomorecharset:
                    self.log.error(
                        u'(P) Single CSSCharsetRule must be before other rules',
                        line, xml.dom.HierarchyRequestErr)
                    continue
                else:
                    rule = self._rules[name](startline=line)
                    try:
                        rule._setCssText(cssText, self)
                    except xml.dom.SyntaxErr, e:
                        self.log.error(u'(P) %s' % e)
                        continue
                    nomorecharset = True

            elif u'CSSImportRule' == name:
                if nomoreimport:
                    self.log.error(
                        u'(P) CSSImportRule must be before other rules except CSSCharsetRule',
                        line, xml.dom.HierarchyRequestErr)
                    continue
                else:
                    nomorecharset = True

                    rule = self._rules[name](startline=line)
                    try:
                        rule._setCssText(cssText, self)
                    except xml.dom.InvalidCharacterErr, e:
                        self.log.warn(u'(P) %s' % e,
                            line, xml.dom.InvalidCharacterErr)
                        continue
                    except xml.dom.SyntaxErr, e:
                        self.log.error(u'(P) %s' % e,
                            line, xml.dom.SyntaxErr)
                        continue

            # anywhere
            elif name == u'Comment':
                rule = self._rules[name](startline=line)
                try:
                    rule._setCssText(cssText, self)
                except xml.dom.SyntaxErr, e:
                    self.log.error(u'(P) %s' % e,
                        line, xml.dom.SyntaxErr)
                    continue                

            # these need to be after @charset and @import
            elif name <> u'CSSUnknownRule': # should be @media or ruleset
                rule = self._rules[name](startline=line)
                try:
                    rule._setCssText(cssText, self)
                except xml.dom.SyntaxErr, e:
                    self.log.error(u'(P) %s' % e,
                        line, xml.dom.SyntaxErr)
                    continue                
                nomorecharset = True
                nomoreimport = True
                
            # UnknownRule might be before @charset or @import
            else:
                rule = self._rules[name](startline=line)
                try:
                    rule._setCssText(cssText, self)
                except xml.dom.InvalidModificationErr, e:
                    self.log.warn(u'(P) %s' % e,
                        line, xml.dom.InvalidModificationErr)
                    continue
                except xml.dom.SyntaxErr, e:
                    self.log.error(u'(P) %s' % e,
                        line, xml.dom.SyntaxErr)
                    continue
                
            if rule:
                rules.append(rule)
        return rules


def _getParser():
    "returns new Parser"
    return Parser()
