"""
CSSCharsetRule implements DOM Level 2 CSS CSSCharsetRule.

TODO:
    - check encoding syntax and not codecs.lookup?
    - parentRule?
    - parentStyleSheet?
"""
__docformat__ = 'restructuredtext'
__all__ = ['CSSCharsetRule']
__version__ = '0.8a5'


import codecs
import re
import xml.dom

import cssrule
import cssutils


class CSSCharsetRule(cssrule.CSSRule):
    """
    The CSSCharsetRule interface represents a @charset rule in a CSS style
    sheet. The value of the encoding attribute does not affect the encoding
    of text data in the DOM objects; this encoding is always UTF-16
    (also in Python?). After a stylesheet is loaded, the value of the
    encoding attribute is the value found in the @charset rule. If there
    was no @charset in the original document, then no CSSCharsetRule is
    created. The value of the encoding attribute may also be used as a hint
    for the encoding used on serialization of the style sheet.

    The value of the @charset rule (and therefore of the CSSCharsetRule)
    may not correspond to the encoding the document actually came in;
    character encoding information e.g. in an HTTP header, has priority
    (see CSS document representation) but this is not reflected in the
    CSSCharsetRule.

    Format: @charset "encoding";
    
    Properties:

    encoding: of type DOMString
        The encoding information used in this @charset rule.    
    """
    
    type = cssrule.CSSRule.CHARSET_RULE 


    def __init__(self, encoding=None, readonly=False, startline=1):
        """
        if readonly allows setting of encoding in constructor only

        encoding:
            a valid character encoding
        readonly:
            defaults to False        
        startline:
            of this rule
        """
        self._startline = startline
        self._readonly = False
        
        if encoding:
            self.encoding = encoding
        else:
            self.encoding = u''

        self._readonly = readonly


    def _getEncoding(self):
        "returns encoding as a string"
        return self._encoding

    def _setEncoding(self, encoding):
        """
        DOMException on setting
        
        - SYNTAX_ERR: (self)
          Raised if the specified encoding value has a syntax error and
          is unparsable.

          Currently only valid Python encodings are allowed!
            
        - NO_MODIFICATION_ALLOWED_ERR: (CSSRule)
          Raised if this encoding rule is readonly.
        """        
        self._checkReadonly()

        if encoding:
            try:
                codecs.lookup(encoding)
            except LookupError:
                raise xml.dom.SyntaxErr(
                    'Unknown (Python) encoding "%s"' % encoding)

        self._encoding = encoding

    encoding = property(_getEncoding, _setEncoding,
        doc="(DOM)The encoding information used in this @charset rule.")


    def getFormatted(self, indent=4, offset=0):
        "returns inherited property cssText prettyprinted"
        if self._encoding:
            cssText = u'@charset "%s";' % self._encoding
            return offset * u' ' + cssText
        else:
            return u''


    def _setCssText(self, cssText, parser=None):
        """
        parser:
            if called from cssparser directly this is Parser instance
        
        DOMException on setting
         
        - SYNTAX_ERR: (L)
          Raised if the specified CSS string value has a syntax error and
          is unparsable.
        - INVALID_MODIFICATION_ERR: (L)
          Raised if the specified CSS string value represents a different
          type of rule than the current one.
        - HIERARCHY_REQUEST_ERR: (P)
          Raised if the rule cannot be inserted at this point in the
          style sheet.
        - NO_MODIFICATION_ALLOWED_ERR: (CSSRule)
          Raised if the rule is readonly.
        """
        super(CSSCharsetRule, self)._setCssText(cssText)

        if parser is None:
            parser = cssutils._getParser()
            
        encodingC, encoding, semiC =\
            parser.parseCSSCharsetRule(cssText, line=self._startline)
        
        # TODO: CharsetRule comments

        self._setEncoding(encoding)

    cssText = property(fget=getFormatted, fset=_setCssText,
        doc="(DOM) The parsable textual representation.")


if __name__ == '__main__':
    c = CSSCharsetRule('utf-8')
    print c.cssText
    c.cssText = u'@charset "ascii";'
    print c.cssText