"""
CSSImportRule implements DOM Level 2 CSS CSSImportRule.

TODO
    - stylesheet: read and parse linked stylesheet
    - parentRule?
    - parentStyleSheet?
"""
__docformat__ = 'restructuredtext'

__all__ = ['CSSImportRule']
__version__ = '0.8a5'


import xml.dom

import cssrule
import cssutils


class CSSImportRule(cssrule.CSSRule):
    """
    Represents an @import rule within a CSS style sheet.  The @import rule
    is used to import style rules from other style sheets.

    Format: @import url(URISTRING) MediaList;

    Properties:
    
    href: of type DOMString, (DOM readonly)
        The location of the style sheet to be imported. The attribute will
        not contain the url(...) specifier around the URI.
    media: of type stylesheets::MediaList, (DOM readonly)
        A list of media types for this rule of type MediaList.
    stylesheet: of type CSSStyleSheet, (DOM readonly)
        The style sheet referred to by this rule. The value of this
        attribute is None if the style sheet has not yet been loaded or if
        it will not be loaded (e.g. if the stylesheet is for a media type
        not supported by the user agent).    
    """

    type = cssrule.CSSRule.IMPORT_RULE 


    def __init__(self,
                 href=u'', mediaText=u'', readonly=False, startline=1):
        """
        href
            The location of the style sheet to be imported.
        media
            A list of media types for which this style sheet may be used\
            as a string
        readonly
            defaults to False        
        startline
            of this rule
        """
        self._startline = startline
        self._readonly = False

        self.href = href

        self.media = cssutils.stylesheets.MediaList(
            mediaText, readonly=readonly, startline=self._startline)

        # TODO: load stylesheet from href automatically?
        # self._styleSheet = None

        self._readonly = readonly
    

    def _getStyleSheet(self):
        """
        returns a CSSStyleSheet or None
        """
        return None

    styleSheet = property(_getStyleSheet,
        doc="The style sheet referred to by this rule.")


    def getFormatted(self, indent=0, offset=0):
        "returns inherited property cssText or empty string"
        if self.href:
            if self.media.mediaText == u'all':
                text = u'@import url(%s);' % (
                self.href)
            else:
                text = u'@import url(%s) %s;' % (
                self.href, self.media.mediaText)
            return offset * u' ' + text
        else:
            return u''

    def _setCssText(self, cssText, parser=None):
        """
        cssText
            textual text to set
        parser
            if called from cssparser directly this is Parser instance

        DOMException on setting
         
        - SYNTAX_ERR: (L)
            Raised if the specified CSS string value has a syntax error and
            is unparsable.
        - INVALID_MODIFICATION_ERR: (L)
            Raised if the specified CSS string value represents a different
            type of rule than the current one.            
        - HIERARCHY_REQUEST_ERR: (parser)
            Raised if the rule cannot be inserted at this point in the
            style sheet.
        - NO_MODIFICATION_ALLOWED_ERR: (super)
            Raised if the rule is readonly.
        """
        super(CSSImportRule, self)._setCssText(cssText)          

        if parser is None:
            parser = cssutils._getParser()
            
        hrefC, href, media =\
            parser.parseCSSImportRule(cssText, line=self._startline)

        self.href = href
        self.media = cssutils.stylesheets.MediaList(
            media, startline=self._startline)

    cssText = property(fget=getFormatted, fset=_setCssText,
        doc="(DOM attribute) The parsable textual representation.")


if __name__ == '__main__':
    c = CSSImportRule(href='1.css', mediaText='handheld, all')
    print c.cssText
    c.cssText = u'@import url(x.css) tv;'
    print c.cssText