"""
CSSStyleDeclaration implements DOM Level 2 CSS CSSStyleDeclaration.

TODO:
    This interface is also used to provide a read-only access to the
    computed values of an element. See also the ViewCSS interface.

    - return computed values and not literal values
    - simplify unit pairs/triples/quadruples
      2px 2px 2px 2px -> 2px for border/padding...
    - normalize compound properties like:
      background: no-repeat left url()  #fff
      -> background: #fff url() no-repeat left                           
"""
__docformat__ = 'restructuredtext'
__all__ = ['CSSStyleDeclaration']
__version__ = '0.8a5'


import xml.dom 

import node

import cssutils.css
import cssutils


class CSSStyleDeclaration(node.Node):
    """
    The CSSStyleDeclaration class represents a single CSS declaration
    block. This class may be used to determine the style properties
    currently set in a block or to set style properties explicitly
    within the block.
    
    While an implementation may not recognize all CSS properties within
    a CSS declaration block, it is expected to provide access to all 
    specified properties in the style sheet through the 
    CSSStyleDeclaration interface. 
    Furthermore, implementations that support a specific level of CSS
    should correctly handle CSS shorthand properties for that level. For
    a further discussion of shorthand properties, see the CSS2Properties
    interface.

    Format: [Property: Value;]* Property: Value?    

    Properties:

    cssText: of type DOMString
        The parsable textual representation of the declaration block
        (excluding the surrounding curly braces). Setting this attribute
        will result in the parsing of the new value and resetting of the
        properties in the declaration block. It also allows the insertion
        of additional properties and their values into the block.
    length: of type unsigned long, readonly
        The number of properties that have been explicitly set in this
        declaration block. The range of valid indices is 0 to length-1
        inclusive.
    parentRule: of type CSSRule, readonly
        The CSS rule that contains this declaration block or None if this
        CSSStyleDeclaration is not attached to a CSSRule.
    """
  

    def __init__(self, parentRule=None, readonly=False, startline=1):
        """
        parentRule
            The CSS rule that contains this declaration block or None if
            this CSSStyleDeclaration is not attached to a CSSRule.
        readonly
            defaults to False        
        startline
            of this rule
        """
        self._readonly = False
        self._startline = startline
        
        self._parentRule = parentRule
        self._properties = []
        
        self._readonly = readonly


    def _checkReadonly(self):
        """
        raises
            xml.dom.NoModificationAllowedErr if rule is readonly
        """
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr(
                u'CSSStyleDeclaration is readonly.')
        

    def getFormatted(self, indent=4, offset=0):
        out = []
        for p in self._properties:
    
            priority = p[u'priority']
            if priority:
                priority = u' %s' % priority
                
            out.append(offset * u' ' + indent * u' ' + u'%s: %s%s;' %
                       (p[u'name'], p[u'value'], priority))
            
        return u'\n'.join(out)

    def _setCssText(self, cssText, parser=None):
        """
        Setting this attribute will result in the parsing of the new value
        and resetting of all the properties in the declaration block
        including the removal or addition of properties.

        DOMException on setting
        
        - SYNTAX_ERR: (self?)
          Raised if the specified CSS string value has a syntax error and
          is unparsable.
          
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this declaration is readonly or a property is
          readonly.
        """
        self._checkReadonly()
        self.literalCssText = cssText # for rules in CSSRule super class   
        if parser is None:
            parser = cssutils._getParser()
            
        properties = parser.parseCSSStyleDeclaration(
            cssText, line=self._startline)
        
        line = self._startline
        for p in properties:
            name, value, priority =\
                parser.parseCSSProperty(
                    p, line=self._startline)
            self.setProperty(
                name, value, priority, startline=line, parser=parser)

            if name:
                line += name.count(u'\n')
            if value:
                line += value.count(u'\n')
            if priority:
                line += priority.count(u'\n')


    cssText = property(getFormatted, _setCssText,
        doc="(DOM) The parsable textual representation of the declaration\
        block excluding the surrounding curly braces.")


    def _getLength(self):
        return len(self._properties)
    
    length = property(_getLength,
        doc="(DOM) the number of properties that have been explicitly set\
        in this declaration block. The range of valid indices is 0 to\
        length-1 inclusive.")


    def _getParentRule(self):
        return self._parentRule
        
    parentRule = property(_getParentRule,
        doc="(DOM) The CSS rule that contains this declaration block or\
        None if this CSSStyleDeclaration is not attached to a CSSRule.")


    def _getPropertyIndex(self, name):
        """
        (cssutils) helper

        name
            of CSS property

        returns the index of the Property with name 
        """
        for i, p in enumerate(self._properties):
            if p[u'name'] == name:
                return i
        return None

    
    # TODO
    def getPropertyCSSValue(self, name):
        """
        (DOM)
        Used to retrieve the object representation of the value of a CSS
        property if it has been explicitly set within this declaration
        block. This method returns None if the property is a shorthand
        property. Shorthand property values can only be accessed and
        modified as strings, using the getPropertyValue and setProperty
        methods.

        name
            of the CSS property
        
        returns CSSValue, the value of the property if it has been
        explicitly set for this declaration block. Returns None if the
        property has not been set.
        """
        # TODO: shorthand properties return None or real value?

        i = self._getPropertyIndex(name)
        if i is not None:
            v = cssutils.css.CSSValue(cssText=self.getPropertyValue(name))
            return v
        else:
            return None


    def getPropertyPriority(self, name):
        """
        (DOM)
        Used to retrieve the priority of a CSS property (e.g. the
        "important" qualifier) if the property has been explicitly set in
        this declaration block.
        
        name
            of the CSS property
        
        returns a string representing the priority (e.g. "important") if
        one exists. The empty string if none exists.
        """
        i = self._getPropertyIndex(name)
        if i is not None:
            return self._properties[i][u'priority']
        else:
            return u''


    def getPropertyValue(self, name):
        """
        (DOM)
        Used to retrieve the value of a CSS property if it has been
        explicitly set within this declaration block.
        
        name
            of the CSS property
        
        returns the value of the property if it has been explicitly set
        for this declaration block. Returns the empty string if the
        property has not been set. 
        """
        i = self._getPropertyIndex(name)
        if i is not None:
            return self._properties[i][u'value']
        else:
            return u''


    def item(self, index):
        """
        (DOM)
        Used to retrieve the properties that have been explicitly set in
        this declaration block. The order of the properties retrieved using
        this method does not have to be the order in which they were set.
        This method can be used to iterate over all properties in this
        declaration block.

        index
            of the property to retrieve
        
        returns the name of the property at this ordinal position. The
        empty string if no property exists at this position.
        """            
        try:
            return self._properties[index][u'name']
        except IndexError:
            return u''


    def removeProperty(self, name):
        """
        (DOM)
        Used to remove a CSS property if it has been explicitly set within
        this declaration block.
        
        name
            of the CSS property to remove
        
        returns the value of the property if it has been explicitly set for
        this declaration block. Returns the empty string if the property
        has not been set or the property name does not correspond to a
        known CSS property
        
        raises DOMException
        
        - NO_MODIFICATION_ALLOWED_ERR: (self, Property)
          Raised if this declaration is readonly or the property is
          readonly.
        """
        self._checkReadonly()
        
        try:
            r = self.getPropertyValue(name)
            del self._properties[self._getPropertyIndex(name)]
        except (TypeError, ValueError):
            r = u''
            
        return r


    def setProperty(self, name, value, priority=u'',
                    startline=1, parser=None):
        """
        (DOM)
        Used to set a property value and priority within this declaration
        block.
        
        name
            of the CSS property to set
            (in W3C DOM the parameter is called "propertyName")
        value
            the new value of the property
        priority
            the optional priority of the property (e.g. "important")
        startline
            in CSS source file
        parser
            optional CSSParser object, used by CSSParser only
            
        DOMException on setting
        
        - SYNTAX_ERR: (lexer)
          Raised if the specified value has a syntax error and is
          unparsable.
          
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this declaration is readonly or the property is
          readonly.
        """
        # TODO: parser might return None?!
        if name is None or value is None:
            return
        
        self._checkReadonly()
        if parser is None:
            parser = cssutils._getParser()
        
        name, nameorder = parser.parsePropertyName(name, line=startline)
        if name is None:
            return
        
        value = parser.parsePropertyValue(name, value, line=startline) #!
        if value is None:
            return

        priority = parser.parsePropertyPriority(priority, line=startline)
                   
        self.removeProperty(name)
        self._properties.append(
            {   u'name': name,
                u'nameorder': nameorder,
                u'value': value,
                u'priority': priority
                })


if __name__ == '__main__':
    d = CSSStyleDeclaration()
    d.cssText = '{color: red ; left: 0 }'
    print d.cssText
    d.cssText = '{color1: red ; left: 0  ;    }'
    print d.cssText
