"""
CSSStyleRule implements DOM Level 2 CSS CSSStyleRule.

TODO
    - selectorText parsing (raise Syntax Error), comments
        CSS2 gives a special meaning to the comma (,) in selectors.
        However, since it is not known if the comma may acquire other
        meanings in future versions of CSS, the whole statement should be
        ignored if there is an error anywhere in the selector, even though
        the rest of the selector may look reasonable in CSS2.

        Illegal example(s):

        For example, since the "&" is not a valid token in a CSS2 selector,
        a CSS2 user agent must ignore the whole second line, and not set
        the color of H3 to red:
        
    - parentRule?
    - parentStyleSheet?
    
    auto box hacks?
    (normalize order of declarations -> CSSNormalizer)

    indent subrules if selector is subselector
        contains(selector)
        isSubselector(selector)
"""
__docformat__ = 'restructuredtext'
__all__ = ['CSSStyleRule']
__version__ = '0.8a5'


import types
import xml.dom

import cssrule
import cssstyledeclaration

import cssutils


class CSSStyleRule(cssrule.CSSRule):
    """
    represents a single rule set in a CSS style sheet.

    Format: selectorText { style block }

    Properties:
    
    selectorText: of type DOMString
        The textual representation of the selector for the rule set. The
        implementation may have stripped out insignificant whitespace while
        parsing the selector.
    style: of type CSSStyleDeclaration, (DOM readonly)
        The declaration-block of this rule set.    
    """

    type = cssrule.CSSRule.STYLE_RULE 


    def __init__(self,
            selectorText=u'', style=None, readonly=False, startline=1):
        """
        selectorText
            type string
        style
            CSSStyleDeclaration of this CSSStyleRule
        readonly
            defaults to False        
        startline
            of this rule
        """
        self._startline = startline
        self._readonly = False
        
        if selectorText:
            self.selectorText = selectorText
        else:
            self._selectors = []        
        self.setStyleDeclaration(style)

        self._readonly = readonly

        
    def addSelector(self, selector):
        """
        (cssutils)
        append a selector to this rule
        
        selector
            a selector of type string, may also contain several single
            selectors seperated by ","

        DOMException on setting
        
        - SYNTAX_ERR:
          Raised if the specified CSS string value has a syntax error and
          is unparsable.     
        - NO_MODIFICATION_ALLOWED_ERR:
          Raised if this rule is readonly.

        TODO: should use set selectorText? or other way round?
        """
        self._checkReadonly()

        # TODO: parse        
        for s in selector.split(u','):
            # norm WS
            self._selectors.append(u' '.join(s.split()))


    def getSelectors(self):
        """
        (cssutils)
        returns a Python list of selector strings
        see selectorText for a single textual representation
        """
        return self._selectors

  
    def _getSelectorText(self):
        """
        returns a comma separated string of selectors
        """
        return u', '.join(self._selectors)

    def _setSelectorText(self, selectorText):
        """
        selector
            of type string, might also be a comma separated list of
            selectors

        DOMException on setting
        
        - SYNTAX_ERR:
          Raised if the specified CSS string value has a syntax error
          and is unparsable.     
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this rule is readonly.
        """
        self._checkReadonly()
        
        if not selectorText:
            raise xml.dom.SyntaxErr('Empty text for selectorText given.')

        self._selectors = [] # empty current
        self.addSelector(selectorText)

    selectorText = property(_getSelectorText, _setSelectorText,
        doc="""(DOM) The textual representation of the selector for the
            rule set.""")


    def getStyleDeclaration(self):
        return self._style

    def setStyleDeclaration(self, style):
        """
        style
            StyleDeclaration or string
        """
        self._checkReadonly()

        if type(style) in types.StringTypes:
            temp = cssstyledeclaration.CSSStyleDeclaration()
            temp.cssText = style
            style = temp
        self._style = style

    style = property(getStyleDeclaration, setStyleDeclaration,
        doc="(DOM readonly) The declaration-block of this rule set.")


    def getFormatted(self, indent=4, offset=0):
        if self.selectorText:
            offsetspace = offset * u' '
            indentspace = indent * u' '
            out = []
            if self._style:
                out.append(offsetspace + self._getSelectorText() + u' {')
                out.append(self._style.getFormatted(indent, offset))
                out.append(offsetspace + indentspace + u'}')
            else:
                out.append(offsetspace + self._getSelectorText() + u' {}\n')
            return u'\n'.join(out)
        else:
            return u''

    def _setCssText(self, cssText, parser=None):
        """
        cssText
            textual text to set
        parser
            if called from cssparser directly this is CSSParser instance

        DOMException on setting
        
        - SYNTAX_ERR: (L, StyleDeclaration)
          Raised if the specified CSS string value has a syntax error and
          is unparsable.
        - INVALID_MODIFICATION_ERR: (L) 
          Raised if the specified CSS string value represents a different
          type of rule than the current one.
        - HIERARCHY_REQUEST_ERR: (P)
          Raised if the rule cannot be inserted at this point in the
          style sheet.
        - NO_MODIFICATION_ALLOWED_ERR: (super)
          Raised if the rule is readonly.
        """
        super(CSSStyleRule, self)._setCssText(cssText)

        if parser is None:
            parser = cssutils._getParser()
            
        selectorText, styledeclarationText = \
            parser.parseCSSStyleRule(
                cssText, line=self._startline)

        # TODO: parse selector
        self.selectorText = selectorText
       
        style = cssstyledeclaration.CSSStyleDeclaration(
            parentRule=self,
            startline=self._startline + selectorText.count(u'\n'))
        style._setCssText(styledeclarationText, parser)
        self.style = style
            
    cssText = property(getFormatted, _setCssText,
        doc="(DOM) The parsable textual representation of the rule.")


if __name__ == '__main__':
    r = CSSStyleRule(selectorText='a')
    r.cssText = 'a, b { color: red }'
    r.addSelector('x')
    print r.getSelectors()
    print r.cssText


