"""
CSSStyleSheet implements DOM Level 2 CSS CSSStyleSheet.

TODO:
    ownerRule and ownerNode
CHECK
    insertRule HIERARCHY_REQUEST_ERR for all cases?
"""
__docformat__ = 'restructuredtext'

__all__ = ['CSSStyleSheet']
__version__ = '0.8a5'


import types
import xml.dom

import cssutils.css
import cssutils.stylesheets
import cssutils


class CSSStyleSheet(cssutils.stylesheets.StyleSheet):
    """
     The CSSStyleSheet interface is a concrete interface used to represent
     a CSS style sheet i.e., a style sheet whose content type is
     "text/css".

    Properties:

    cssRules: of type CSSRuleList, (DOM readonly)
        The list of all CSS rules contained within the style sheet. This
        includes both rule sets and at-rules.
    ownerRule: of type CSSRule, readonly
        If this style sheet comes from an @import rule, the ownerRule
        attribute will contain the CSSImportRule. In that case, the
        ownerNode attribute in the StyleSheet interface will be None. If
        the style sheet comes from an element or a processing instruction,
        the ownerRule attribute will be None and the ownerNode attribute
        will contain the Node.

    Inherits properties from stylesheet.StyleSheet!
    """
    
    def __init__(self, type='text/css', href=None, media=None,
                 title='', disabled=None,
                 ownerNode=None, parentStyleSheet=None,
                 readonly=False):
        
        super(CSSStyleSheet, self).__init__(
                type, href, media, title, disabled,
                ownerNode, parentStyleSheet)

        self._readonly = readonly
        self.cssRules = cssutils.css.CSSRuleList()


    def _checkReadonly(self):
        """
        raises xml.dom.NoModificationAllowedErr if rule is readonly
        """
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr(
                u'CSSStyleSheet is readonly.')


    def pprint(self, indent=4):
        """
        (cssutils)
        pretty prints the complete StyleSheet.

        indent:
            with this number of spaces
        """
        print self._getCssText(indent)


    def _getCssText(self, indent=4):
        out = []
        for rule in self.cssRules:
            out.append(rule.getFormatted(indent))
        return '\n'.join(out)

    def _setCssText(self, cssText, parser=None):
        """
        (cssutils)
        Parses cssText and overwrites the whole stylesheet.

        cssText
            textual text to set
        parser
            if called from cssparser directly this is CSSParser instance

        DOMException on setting
         
        - SYNTAX_ERR:
            Raised if the specified CSS string value has a syntax error and
            is unparsable.
            
        - NO_MODIFICATION_ALLOWED_ERR: (self)
            Raised if the rule is readonly.        
        """
        self._checkReadonly()

        self.literalCssText = cssText        

        if parser is None:
            parser = cssutils._getParser()        

        self.cssRules = parser.parseCSSStyleSheet(cssText)

    cssText = property(_getCssText, _setCssText,
            "(cssutils) a textual representation of the stylesheet")


    def deleteRule(self, index):
        """
        Used to delete a rule from the style sheet.

        index
            of the rule to remove in the StyleSheet's rule list  

        DOMException
        
        - INDEX_SIZE_ERR: (self)
            Raised if the specified index does not correspond to a rule in
            the style sheet's rule list.
        - NO_MODIFICATION_ALLOWED_ERR: (self)
            Raised if this style sheet is readonly.
        """
        self._checkReadonly()
        
        try:
            del self.cssRules[index]
        except IndexError:
            raise xml.dom.IndexSizeErr(
                u'Index %s not in the list with len %s' % (
                    index, len(self.cssRules)))


    def addRule(self, rule):
        """
        (cssutils)
        Append a Rule to the StyleSheet, see insertRule()

        rule
            the Rule object or string
        """
        self.insertRule(rule, len(self.cssRules))


    def insertRule(self, rule, index):
        """
        Used to insert a new Rule **string** into the
        style sheet. The new rule now becomes part of the cascade.

        rule
            a parsable DOMString (cssutils: or Rule object)
        index
            of the rule before the new rule will be inserted.
            If the specified index is equal to the length of the
            StyleSheet's rule collection, the rule will be added to the end
            of the style sheet.
            
        returns the index within the stylesheet's rule collection

        DOMException
        
        - HIERARCHY_REQUEST_ERR: (self)
          Raised if the rule cannot be inserted at the specified index
          e.g. if an @import rule is inserted after a standard rule set
          or other at-rule.
        - INDEX_SIZE_ERR: (self)
          Raised if the specified index is not a valid insertion point.
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this style sheet is readonly.
        - SYNTAX_ERR: (P)
          Raised if the specified rule has a syntax error and is
          unparsable.        
        """
        self._checkReadonly()

        # check position
        if index < 0:
            index = 0
        if index > len(self.cssRules):
            raise xml.dom.IndexSizeErr(
                u'index %s higher than current list length, len is only %s.'
                % ( index, len(self.cssRules)))

        # parse
        if type(rule) in types.StringTypes:
            stylesheet = cssutils.css.CSSStyleSheet()
            stylesheet.cssText = rule
            if len(stylesheet.cssRules) != 1:
                raise xml.dom.SyntaxErr(
                    u'Invalid Rule: %s' % rule)
            rule = stylesheet.cssRules[0]

        # check order (@import must be 1st)
        if cssutils.css.CSSRule.IMPORT_RULE == rule.type:
            for n in self.cssRules[:index]:
                # comments have to have a type?
                if n.type \
                 and n.type != cssutils.css.CSSRule.IMPORT_RULE \
                 and n.type != cssutils.css.CSSRule.CHARSET_RULE:
                    raise xml.dom.HierarchyRequestErr(
                        'ImportRule must be before other rules (except CharsetRule)')

        # insert
        self.cssRules.insert(index, rule)
        return index

            
    def _getsetOwnerRuleDummy(self):
        """
        NOT IMPLEMENTED YET
        If this style sheet comes from an @import rule, the ownerRule
        attribute will contain the CSSImportRule. In that case, the
        ownerNode attribute in the StyleSheet interface will be null. If
        the style sheet comes from an element or a processing instruction,
        the ownerRule attribute will be null and the ownerNode attribute
        will contain the Node.
        """
        raise NotImplementedError()

    ownerRule = property(_getsetOwnerRuleDummy, _getsetOwnerRuleDummy,
                         doc="(DOM attribute) NOT IMPLEMENTED YET")


if __name__ == '__main__':
    print "CSSStyleSheet"
    c = CSSStyleSheet(type='text/css',
                        href=None,
                        #
                        media="all",
                        title="test",
                        disabled=None,
                        ownerNode=None,
                        parentStyleSheet=None)
    c.cssText = 'a { color: red }'
    print "title: ", c.title
    print "type: ", c.type

    print "cssText: ", c.cssText

    print c.cssRules    
    print c.pprint(8)
    