#!/usr/bin/env python
"""
Some basic helper functions to deal with encodings of text files
(like HTML, XHTML, XML) via HTTP.

====================
      encutils
====================
:Author: Christof Hoeke
:Date: 050820
:License: This work is licensed under a
    Creative Commons License http://creativecommons.org/licenses/by/2.0/

Website
    http://cthedot.de/encutils/

references
----------
HTML
    http://www.w3.org/TR/REC-html40/charset.html#h-5.2.2
XML
    http://www.xml.com/pub/a/2004/07/21/dive.html

TODO:
    - XML autodetection
    - parse with re as content_type might be more complex...
    - use mimetools.Message for parsing Meta info???
    - HTML meta elements in comments? (remove comments?)

history
-------
v0.4.1 050820
    docs adjusted to epydoc automatic doc generation with ReST
    
v0.4 050817
    - added getXMLEncoding(text) which finds explicit encoding
        TODO: should autodetect XML encoding? maybe with parameter only?
    - guessEncoding should work as defined in specifications
    - constants _NON_XML_TYPE, _APPLICATION_TYPE, _TEXT_TYPE for different
        types of served XML files, internal use and probably not useful to
        users of encutils
    - cleaned up code

v0.3 050730
    - _splitContentTypeValue returns encoding if given uppercase now
    - running standalone shows help now (is there a simpler way as below?)

v0.2 050703
    cleaned up documentation and API

v0.1 050626
    first release     
"""
__docformat__ = 'restructuredtext'

__all__ = ['encodingByMediaType',
           'getHTTPInfo',
           'getMetaInfo',
           'getXMLEncoding',
           'guessEncoding']

__version__ = '0.4.1'


import re
import types


_NON_XML_TYPE = 0
# files server e.g. as text/html, text/plain etc
_APPLICATION_TYPE = 1
# files served as application/...
_TEXT_TYPE = 2
# files served as text/...


def _splitContentTypeValue(content_type, log=None):
    """
    Returns (media_type, encoding) tuple by splitting Content_Type
    HTTP header or meta element string. Encoding might
    be None if no charset given.

    example:
        content_type = "text/html; charset=utf-8"
        returns "text/html", "UTF-8"
    """
##    try:
##        encoding = re.findall(ur'charset=(.*)', content_type,
##               re.DOTALL|re.IGNORECASE|re.UNICODE)[0]
##    except IndexError:
##        encoding = None
        
    if content_type.find(u';') <> -1:    
        media_type, encoding = content_type.split(u';')
        media_type = media_type.strip()
        if encoding:
            charseteq, encoding = encoding.split(u'=')
            encoding = encoding.strip().upper()
    else:
        media_type, encoding = content_type.strip(), None
        
    if log:
        log.debug(u'_splitContentTypeValue')
        log.debug(u'    input : "%s" ', content_type)
        log.debug(u'    output: "%s", "%s"', media_type, encoding)
    
    return media_type, encoding


def _getXMLType(media_type, log=None):
    """
    returns
        _NON_XML_TYPE = for unknown types
        _APPLICATION_TYPE = for XML types served as application/...
        _TEXT_TYPE = for XML/HTML types served as text/...
    """
    type1 = [
        ur'application/.*?\+xml',
        u'application/xml',
        u'application/xml-dtd',
        u'application/xml-external-parsed-entity']
    type2 = [
        ur'text\/.*?\+xml',
        u'text/xml',
        u'text/xml-external-parsed-entity']
    
    media_type = media_type.strip().lower()

    xmltype = _NON_XML_TYPE
    if media_type in type1 or\
            re.match(type1[0], media_type, re.I|re.S|re.X):
        xmltype = _APPLICATION_TYPE
    if media_type in type2 or\
            re.match(type2[0], media_type, re.I|re.S|re.X):
        xmltype = _TEXT_TYPE

    if log:
        log.debug(u'_getXMLType')
        log.debug(u'    media_type : "%s" ', media_type)
        log.debug(u'    xmltype    : %s', xmltype)
        
    return xmltype


def encodingByMediaType(media_type, log=None):
    """
    Returns a default encoding for the given
    Media-Type, e.g. ``'UTF-8'`` for ``media_type='application/xml'``. 
   
    If no default encoding is available returns ``None``.
    """
    encodings = {
        _APPLICATION_TYPE: u'UTF-8',
        _TEXT_TYPE: u'ASCII',
        _NON_XML_TYPE: u'ASCII'}

    xmltype = _getXMLType(media_type)
    encoding = encodings[xmltype]

    if log:
        log.debug(u'encodingByMediaType')
        if not encoding:
            log.info(u'(encutils) "%s" Media-Type has no default encoding',
                media_type)
        else:
            log.info(
                u'(encutils) Default encoding for Media Type "%s" : %s',
                media_type, encoding)

    return encoding


def getHTTPInfo(response, log=None):
    """
    Returns ``(media_type, encoding)`` information from the response'
    Content-Type HTTP header (Case of headers is ignored.)
    May be ``(None, None)`` e.g. if no Content-Type header is
    available.
    """
    info = response.info()
    media_type = info.gettype()
    encoding = info.getparam('charset') 

    if encoding:
        encoding = encoding.upper()
        if log:
            log.info(u'(encutils) HTTP media_type: %s', media_type)
            log.info(u'(encutils) HTTP encoding  : %s', encoding)
    return media_type, encoding


def getMetaInfo(text, log=None):
    """
    Returns (media_type, encoding) information from (first)
    X/HTML Content-Type ``<meta>`` element if available.

    Normally in X/HTML:
        ``<meta http-equiv="Content-Type" content="media_type;
        charset=encoding"/>``
    """
    ctmetas = re.findall(ur'''<meta.*?
            http-equiv\s* = \s*['"]\s*Content-Type\s*['"]\s*
            .*?\/?>
        ''', text, re.I|re.S|re.U|re.X)

    if ctmetas:
        first = ctmetas[0]
        value = re.findall(ur'''
                content\s*=\s*  # content= 
                ['"]\s*         # " or '
                (.*?)           # find only value text
                \s*['"]         # " or '
            '''
            ,first, re.I|re.S|re.U|re.X)
        if value:
            media_type, encoding = _splitContentTypeValue(value[0], log)
            if log:
                log.info(u'(encutils) HTML <meta>   : %s', value[0])
                log.info(u'(encutils)     media_type: %s', media_type)
                log.info(u'(encutils)     encoding  : %s', encoding)
            # never default encoding for HTML
    else:
        media_type = encoding = None

    return media_type, encoding


def getXMLEncoding(text, log=None):
    """
    Parses XML declaration of a document (if present) (simplified).
    Returns ``(encoding, explicit)``.
    No autodetection of BOM is done yet. If no explicit encoding is found
    returns ``('UTF-8', False)``.
    """
    reenc = re.findall(ur'''^.*? encoding=['"](.*?)['"] .*? \?>''',
                       text, re.I|re.S|re.U|re.X)
    try:
        encoding = reenc[0]
    except IndexError:
        encoding = u'UTF-8'
        explicit = False
    else:
        explicit = True
    if log:
        log.info(u'(encutils) XML encoding: %s (explicitly set: %s)' %
                 (encoding, explicit))
        
    return encoding.upper(), explicit


def guessEncoding(response, text, log=None):
    """
    - response: response of e.g. urllib.urlopen('url')
    - text: to guess encoding for, might include meta information or XML
      prolog with encoding attribute
    - log: an optional logging logger to which messages may go

    Tries to find the encoding of given text. Uses information in headers
    of supplied HTTPResponse, possible XML declaration and X/HTML
    ``<meta>`` elements.
    
    Returns ``(encoding, mismatch)``.
        - Encoding is the explicit or implicit encoding or None and
          returned always uppercase.
        - Mismatch is True if any mismatches between media_type,
          XML declaration or textcontent are found. More detailed mismatch
          reports are written to the optional log.

        Mismatches are not nessecarily errors!
        For details see the specifications.
    

    RFC 3023 states if media type given in the Content-Type HTTP header is
    application/xml, application/xml-dtd,
    application/xml-external-parsed-entity, or any one of the subtypes of
    application/xml such as application/atom+xml or application/rss+xml
    etc then the character encoding is determined in this order:
    1. the encoding given in the charset parameter of the Content-Type HTTP
    header, or
    2. the encoding given in the encoding attribute of the XML declaration
    within the document, or
    3. utf-8.

    Mismatch possibilities:
        - HTTP + XMLdecla 
        - HTTP + HTMLmeta

        application/xhtml+xml
            XMLdecla + HTMLmeta   

        
    If the media type given in the Content-Type HTTP header is text/xml,
    text/xml-external-parsed-entity, or a subtype like text/Anything+xml,
    the encoding attribute of the XML declaration is ignored completely
    and the character encoding is determined in the order:
    1. the encoding given in the charset parameter of the Content-Type HTTP
    header, or
    2. ascii.

    Mismatch possibilities:
        - HTTP + XMLdecla 
        - HTTP + HTMLmeta 
        
        text/xhtml+xml
            XMLdecla + HTMLmeta 


    For HTML served as text/html:
    1. the encoding given in the charset parameter of the Content-Type HTTP
    header, or
    2. meta encoding
    2. ascii.

    Mismatch possibilities:
        - HTTP + HTMLmeta 
    """
    header_media_type, header_encoding = getHTTPInfo(response, log)
    xmltype = _getXMLType(header_media_type, log)
    xml_encoding, explicit = getXMLEncoding(text, log)
    meta_media_type, meta_encoding = getMetaInfo(text, log)

    if log:
        log.debug('xmltype        : %s' % xmltype)
        log.debug('header_encoding: %s' % header_encoding)
        log.debug('xml_encoding   : %s' % xml_encoding)
        log.debug('meta_encoding  : %s' % meta_encoding)

    mismatch = False
    
    # 1. try Content Type HTTP header
    encoding = header_encoding

    # XML application/...
    if xmltype == _APPLICATION_TYPE:
        if not encoding:
            encoding = xml_encoding # defaults to UTF-8
            
        # possible mismatch
        if header_encoding and xml_encoding <> u'UTF-8' and\
           header_encoding <> xml_encoding:
            mismatch = True
            if log:
                log.warning(u'(encutils) HTTP header <> XML declaration encoding mismatch, using HTTP header encoding')
                log.warning(u'(encutils) \tHTTP "%s" <> XML "%s"',
                    header_encoding, xml_encoding)
        elif header_encoding and meta_encoding and\
             header_encoding <> meta_encoding:
            mismatch = True
            if log:
                log.info(u'(encutils) HTTP header <> XHTML <meta> encoding mismatch, using HTTP header encoding')
                log.info(u'(encutils) \tHTTP "%s" <> META "%s"',
                    header_encoding, meta_encoding)
        elif xml_encoding <> u'UTF-8' and meta_encoding and\
             xml_encoding <> meta_encoding:
            mismatch = True
            if log:
                log.info(u'(encutils) XML declaration encoding <> XHTML <meta> encoding mismatch')
                log.info(u'(encutils) \tXML "%s" <> META "%s"',
                    xml_encoding, meta_encoding)

    # XML text/...
    elif xmltype == _TEXT_TYPE:
        if not encoding:
            encoding = u'ASCII'

        # possible mismatch
        if header_encoding and xml_encoding <> u'UTF-8' and\
           header_encoding <> xml_encoding:
            mismatch = True
            if log:
                log.info(u'(encutils) HTTP header <> XML declaration encoding mismatch, using HTTP header encoding')
                log.info(u'(encutils) \tHTTP "%s" <> XML "%s"',
                    header_encoding, xml_encoding)
        elif header_encoding and meta_encoding and\
             header_encoding <> meta_encoding:
            mismatch = True
            if log:
                log.info(u'(encutils) HTTP header <> XHTML <meta> encoding mismatch, using HTTP header encoding')
                log.info(u'(encutils) \tHTTP "%s" <> META "%s"',
                    header_encoding, meta_encoding)
        elif xml_encoding <> u'UTF-8' and meta_encoding and\
             xml_encoding <> meta_encoding:
            mismatch = True
            if log:
                log.info(u'(encutils) XML declaration encoding <> XHTML <meta> encoding mismatch')
                log.info(u'(encutils) \tXML "%s" <> META "%s"',
                    xml_encoding, meta_encoding)

    # other
    else:
        if not encoding:
            encoding = meta_encoding
            if not encoding:
                encoding = u'ASCII'

        # possible mismatch
        if header_encoding and meta_encoding and\
           header_encoding <> meta_encoding\
           :
            mismatch = True
            if log:
                log.warning(u'(encutils) HTTP header <> HTML <meta> encoding mismatch, using HTTP header encoding')
                log.warning(u'(encutils) \tHTTP "%s" <> META "%s"',
                    header_encoding, meta_encoding)
           
    if log:
        log.info(u'(encutils) Encoding guessed: %s (Mismatch: %s)',
                 encoding, mismatch)

    return encoding, mismatch


if __name__ == '__main__':
    print __doc__
    print "\nFUNCTIONS:\n"
    for x in __all__:
        print x
        exec "print " + x + ".__doc__"
        
    
        
