"""
MediaList implements DOM Level 2 Style Sheets MediaList.

TODO:
    delete: maybe if deleting from all, replace *all* with all others?
    is unknown media an exception?
"""
__docformat__ = 'restructuredtext'
__all__ = ['MediaList']
__version__ = '0.8a5'


import xml.dom

import cssutils


class MediaList(list):
    """
    Provides the abstraction of an ordered collection of media,
    without defining or constraining how this collection is
    implemented. 
    An empty list is the same as a list that contains the medium "all".
    
    If "all" is in the list, every other media *except* "handheld" will be 
    stripped. This is because how Opera handles CSS for PDAs.
    """
    
    _MEDIA = [u'all', u'aural', u'braille', u'embossed', u'handheld',
        u'print', u'projection', u'screen', u'tty', u'tv']
    "possible media types"
    

    def __init__(self, mediaText=None, readonly=False, startline=1):
        """
        mediaText
            unicodestring of media
        startline
            of this rule
        """
        self._startline = startline
        self._readonly = False
        
        if mediaText:
            self._setMediaText(mediaText)

        self._readonly = readonly


    def _checkReadonly(self):
        "raises xml.dom.NoModificationAllowedErr if rule is readonly"
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr(
                u'MediaList is readonly.')


    def _getLength(self):
        "same as Python len(MediaListInstance)"
        return len(self)

    length = property(_getLength,
        doc="(DOM readonly) The number of media in the list.")


    def getFormatted(self):
        "comma-separated list of media, default is 'all'"
        if len(self) == 0:
            return u'all'
        else:
            return u', '.join(self)

    def _setMediaText(self, mediaText, parser=None):
        """
        mediaText
            simple value or comma-separated list of media

        DOMException
        
        - SYNTAX_ERR: (lexer)
            Raised if the specified string value has a syntax error and is
            unparsable.
        - NO_MODIFICATION_ALLOWED_ERR: (self)
            Raised if this media list is readonly.            
        """
        self._checkReadonly()

        if parser is None:
            parser = cssutils._getParser()
            
        # media and comment nodes
        media = parser.parseMediaList(mediaText, line=self._startline)

        del self[:]
        for m in media:
            self.appendMedium(m)

    mediaText = property(getFormatted, _setMediaText,
        doc="""(DOM) The parsable textual representation of the media list.
            This is a comma-separated list of media.""")    

    
    def appendMedium(self, newMedium):
        """
        (DOM)
        Adds the medium newMedium to the end of the list. If the newMedium
        is already used, it is first removed.

        DOMException
        
        - INVALID_CHARACTER_ERR: (self) ?
          If the medium contains characters that are invalid in the
          underlying style language.
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this list is readonly.         
        """
        self._checkReadonly()

        # ? should check format only?
        if newMedium not in self._MEDIA:
            raise xml.dom.InvalidCharacterErr(
                u'"%s" is not a valid medium' % newMedium)

        # all contains every other (except handheld!)
        if u'all' in self and newMedium != u'handheld':
            return

        if newMedium == u'all':
            if u'handheld' in self:
                del self[:]
                self.append(u'all')
                self.append(u'handheld')
            else:
                del self[:]
                self.append(u'all')
            return
        else:
            if newMedium in self:
                self.remove(newMedium)
            self.append(newMedium)

    
    def deleteMedium(self, oldMedium):
        """
        (DOM)
        Deletes the medium indicated by oldMedium from the list.

        DOMException
        
        - NO_MODIFICATION_ALLOWED_ERR: (self)
          Raised if this list is readonly.
        - NOT_FOUND_ERR: (self)
          Raised if oldMedium is not in the list.         
        """
        self._checkReadonly()
        try:
            self.remove(oldMedium)
        except ValueError:
            raise xml.dom.NotFoundErr(
                u'"%s" not in this MediaList' % oldMedium)


    def item(self, index):
        """
        (DOM)
        Returns the indexth in the list. If index is greater than or equal
        to the number of media in the list, this returns None.
        """
        try:
            return self[index]
        except IndexError:
            return None


if __name__ == '__main__':
    m = MediaList()